# Manage debian services.  Start/stop is the same as InitSvc, but enable/disable
# is special.
Puppet::Type.type(:service).provide :debian, :parent => :init do
  desc <<-EOT
    Debian's form of `init`-style management.

    The only differences from `init` are support for enabling and disabling
    services via `update-rc.d` and the ability to determine enabled status via
    `invoke-rc.d`.

  EOT

  commands :update_rc => "/usr/sbin/update-rc.d"
  # note this isn't being used as a command until
  # http://projects.reductivelabs.com/issues/2538
  # is resolved.
  commands :invoke_rc => "/usr/sbin/invoke-rc.d"
  optional_commands :systemctl => "/bin/systemctl"

  # This isn't being used directly, it's just here to ensure
  # that the /usr/sbin/service binary is available.
  SERVICE = "/usr/sbin/service"
  commands :service_cmd => SERVICE

  defaultfor :operatingsystem => :debian

  def self.runs_on_systemd?
    Dir.exists? "/run/systemd/system"
  end

  def is_sysv_unit?
    # The sysv generator sets the SourcePath attribute to the name of the
    # initscript. Use this to detect whether a unit is backed by an initscript
    # or not.
    source = systemctl(:show, "-pSourcePath", @resource[:name])
    source.start_with? "SourcePath=/etc/init.d/"
  end

  def self.instances
    # We need to merge services with systemd unit files with those only having
    # an initscript. Note that we could use `systemctl --all` to get sysv
    # services as well, however it would only output *enabled* services.
    i = {}
    if self.runs_on_systemd?
      begin
        output = systemctl('list-unit-files', '--type', 'service', '--full', '--all',  '--no-pager')
        output.scan(/^(\S+)\.service\s+(disabled|enabled)\s*$/i).each do |m|
          i[m[0]] = new(:name => m[0])
        end
      rescue Puppet::ExecutionFailure
      end
    end
    get_services(defpath).each do |sysv|
      unless i.has_key?(sysv.name)
        i[sysv.name] = sysv
      end
    end
    return i.values
  end

  # Remove the symlinks
  def disable
    if self.class.runs_on_systemd?
      systemctl(:disable, @resource[:name])
    else
      update_rc @resource[:name], "disable"
    end
  end

  def enabled?
    # Initscript-backed services have no enabled status in systemd, so we
    # need to query them using invoke-rc.d.
    if self.class.runs_on_systemd? and not is_sysv_unit?
      begin
        systemctl("is-enabled", @resource[:name])
        return :true
      rescue Puppet::ExecutionFailure
        return :false
      end
    else
      # TODO: Replace system call when Puppet::Util::Execution.execute gives us a way
      # to determine exit status.  http://projects.reductivelabs.com/issues/2538
      system("/usr/sbin/invoke-rc.d", "--quiet", "--query", @resource[:name], "start")

      # 104 is the exit status when you query start an enabled service.
      # 106 is the exit status when the policy layer supplies a fallback action
      # See x-man-page://invoke-rc.d
      if [104, 106].include?($CHILD_STATUS.exitstatus)
        return :true
      elsif [105].include?($CHILD_STATUS.exitstatus)
        # 105 is unknown, which generally means the iniscript does not support query
        # The debian policy states that the initscript should support methods of query
        # For those that do not, peform the checks manually
        # http://www.debian.org/doc/debian-policy/ch-opersys.html
        if get_start_link_count >= 4
          return :true
        else
          return :false
        end
      else
        return :false
      end
    end
  end

  def get_start_link_count
    Dir.glob("/etc/rc*.d/S??#{@resource[:name]}").length
  end

  def enable
    if self.class.runs_on_systemd?
      systemctl(:enable, @resource[:name])
    else
      update_rc @resource[:name], "defaults"
      update_rc @resource[:name], "enable"
    end
  end

  # The start, stop, restart and status command use service
  # this makes sure that these commands work with whatever init
  # system is installed
  def startcmd
    [SERVICE, @resource[:name], :start]
  end

  # The stop command is just the init script with 'stop'.
  def stopcmd
    [SERVICE, @resource[:name], :stop]
  end

  def restartcmd
    (@resource[:hasrestart] == :true) && [SERVICE, @resource[:name], :restart]
  end

  # If it was specified that the init script has a 'status' command, then
  # we just return that; otherwise, we return false, which causes it to
  # fallback to other mechanisms.
  def statuscmd
    (@resource[:hasstatus] == :true) && [SERVICE, @resource[:name], :status]
  end

end
